@@ -4,9 +4,7 @@
 from __future__ import (absolute_import, division, print_function,
                         unicode_literals)
 
-import sys
-
 from .main import main
 
 if __name__ == '__main__':
-    sys.exit(main())
+    main()
